from ._preprocessing import select_genepair_grn
from ._preprocessing import integration_grn
from ._preprocessing import RNAseq_analysis
from ._preprocessing import RNAseq_analysis_uni
from ._preprocessing import selection_GRNandRNAseq
from ._preprocessing import comparison


from ._plotting import landscape
from ._plotting import ROC_plot
from ._plotting import network_umap_pre
from ._plotting import network_pre
from ._plotting import burst_pre
from ._plotting import barplot
from ._plotting import gene_roles_heatmap
from ._plotting import subnetwork_pre
from ._plotting import scatter_plot
from ._plotting import box_plot3
from ._plotting import fit_3d
from ._plotting import scatter_2d
from ._plotting import box_plot4_2
from ._plotting import heatmap_regulation
from ._plotting import barplot_regulation
from ._plotting import visualize_go_bar
from ._synthetic_data import SSA_coexpression
from ._synthetic_data import gibbs_sample


# -*- coding: utf-8 -*-
import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import rpy2.robjects as ro


def network_visualization(counts_file, gene_interactions_file, burst_info_file, degree_data_file, network_figure, neighbors, min_dist):
    """
    Visualization for gene regulatory interactions network embedded with bursting.
    """
    ro.globalenv['counts_file'] = counts_file
    ro.globalenv['gene_interactions_file'] = gene_interactions_file
    ro.globalenv['burst_info_file'] = burst_info_file
    ro.globalenv['degree_data_file'] = degree_data_file
    ro.globalenv['network_figure'] = network_figure
    ro.globalenv['neighbors'] = neighbors
    ro.globalenv['min_dist'] = min_dist
    network_visualization_R = """
    library(umap)
    library(ggplot2)
    library(ggraph)
    library(igraph)
    library(tidyverse)
    library(RColorBrewer)
    library(readr)
    library(ggrepel)
    library(factoextra)

    set.seed(123)
    expression_counts_matrix_ <- read_csv(counts_file)
    expression_counts_matrix <- expression_counts_matrix_[ ,3:ncol(expression_counts_matrix_)]
    expression_counts_matrix <- as.data.frame(lapply(expression_counts_matrix, function(x) as.numeric(as.character(x))))
    umap_config <- umap.defaults
    umap_config$n_neighbors <- neighbors
    umap_config$min_dist <- min_dist
    umap_result <- umap(expression_counts_matrix, config = umap_config)
    umap_df <- as.data.frame(umap_result$layout)
    colnames(umap_df) <- c("UMAP1", "UMAP2")
    umap_df$Gene <- rownames(expression_counts_matrix)
    # Elbow to confirm cluster number
    # fviz_nbclust(umap_df[, c("UMAP1", "UMAP2")], kmeans, method = "wss") + ggtitle("Elbow Method for Determining Optimal Clusters")
    k <- 3
    kmeans_result <- kmeans(umap_df[, c("UMAP1", "UMAP2")], centers = k)
    umap_df$Cluster <- as.factor(kmeans_result$cluster)

    gene_interacions <- read_csv(gene_interactions_file)
    burst_info_ <- read_csv(burst_info_file)
    burst_info <- data.frame(Gene = umap_df[ ,3], bs = burst_info_[, 3], bf = log10(burst_info_[, 2]))
    network_data <- data.frame(from = gene_interacions[ ,2], to = gene_interacions[ ,3])
    umap_data <- data.frame(Gene = umap_df[ ,3], UMPA1 = umap_df[ ,1], UMPA2 = umap_df[ ,2])
    umap_data <- umap_data %>% left_join(burst_info, by = "Gene")
    graph <- graph_from_data_frame(d = network_data, vertices = umap_data, directed = TRUE)
    degree_data <- data.frame(indegree = degree(graph, mode = "in"), outdegree = degree(graph, mode = "out"))
    # write.csv(degree_data, file = degree_data_file, row.names = FALSE)
    V(graph)$UMAP1 <- umap_df$UMAP1
    V(graph)$UMAP2 <- umap_df$UMAP2
    V(graph)$Size <- umap_data$X1
    V(graph)$Color <- umap_data$X0
    p <- ggraph(graph, layout = "manual", x = V(graph)$UMAP1, y = V(graph)$UMAP2) +
        geom_edge_fan(alpha = 0.025, color = "gray") +
        geom_node_point(aes(size = V(graph)$Size, color = Color), alpha = 1, shape = 16) +
        scale_size_continuous(name = "Burst size", range = c(3, 8), guide = guide_legend(ticks = FALSE, label = FALSE)) + 
        scale_color_viridis_c(name = "Burst frequency", guide = guide_colorbar(ticks = FALSE, label = FALSE))+
        ggtitle("Gene regulatory interactions network embedded with bursting") +
        theme_void() +
        theme(panel.background = element_rect(fill = "white", color = NA), 
              plot.background = element_rect(fill = "white", color = NA),
              plot.title = element_text(hjust = 0.5, size = 15, face = "bold"))
    ggsave(network_figure, plot = p, width = 8, height = 6, dpi = 300)
    """
    ro.r(network_visualization_R)
    return

def TFs_interactiontype_network(readfile_name):
    """
    Visualization for gene regulatory interactions type stacked row chart.
    """
    filtered_inference_result = filter_inference_results(readfile_name, [-3.5, 2.5, -2.5, 5.0])
    feedback_matrix__ = filtered_inference_result[:, np.array([0, 1, 23, 22])]
    unique_genename = sorted(list(set(np.asarray(np.vstack([filtered_inference_result[:, 0], filtered_inference_result[:, 1]])).flatten())))     
    feedback_matrix = np.zeros([len(unique_genename), len(unique_genename)]) - 10
    index = 0
    for genename in unique_genename:
        indices_ = np.where(feedback_matrix__[:, 0] == genename)[0]
        indices__ = np.where(np.isin(unique_genename, feedback_matrix__[indices_, 1]))[0]
        feedback_matrix[indices__, index] = 5 * np.array(feedback_matrix__[indices_, 2].astype(float)).flatten()
        feedback_matrix[index, indices__] = 5 * np.array(feedback_matrix__[indices_, 3].astype(float)).flatten()
        feedback_matrix[index, index] = - 10
        index = index + 1
    feedback_matrix_new = feedback_matrix.copy()
    for index in np.arange(len(unique_genename)):
        feedback_matrix_new[index, :] = sorted(feedback_matrix[index, :], reverse=True)
    feedback_matrix_new = np.hstack([np.arange(len(unique_genename)).reshape([len(unique_genename), 1]), feedback_matrix_new])
    sorted_feedback_matrix =  np.matrix(sorted(feedback_matrix_new.tolist(), key=lambda x: tuple(x[1::])))
    sorted_feedback_matrix_ = np.flipud(sorted_feedback_matrix)
    plt.figure(dpi=300, figsize=(2, 1.5))
    sns.heatmap(sorted_feedback_matrix_[:, 1::], annot=False, cmap=plt.cm.Blues, annot_kws={'size': 100}, linewidths=0.0001, linecolor='w', cbar=False)
    plt.xticks([], [])
    plt.yticks([], [])
    plt.xlabel('Regualtory interaction type', fontsize=3)
    plt.ylabel('TF gene', fontsize=3) 
    plt.title('Regualtory interaction type stacked row chart', fontsize=3)
    plt.show()
    return

def scatterplot_burst(readfile_name, thresholds, group, xlim, ylim):
    """
    Visualization for transcriptional bursting scatter plot.
    """
    filtered_inference_result = filter_inference_results(readfile_name, thresholds)
    bs = np.array(np.vstack([filtered_inference_result[:, 13], filtered_inference_result[:, 14]]).astype(float))
    bf = np.array(np.vstack([filtered_inference_result[:, 11], filtered_inference_result[:, 12]]).astype(float))
    cv2 = np.array(np.vstack([filtered_inference_result[:, 15], filtered_inference_result[:, 16]]).astype(float))
    expression_level = bs * bf

    fig, ax = plt.subplots(dpi=300)
    sc = plt.scatter(np.log10(bf), np.log10(bs), s=cv2*1.5, c=np.log10(expression_level), cmap=plt.cm.viridis)
    cbar = plt.colorbar(sc)
    cbar.ax.tick_params(labelsize=3, width=0.35) 
    cbar.outline.set_linewidth(0.5)
    plt.xlabel('log10(bf)', fontsize=5)
    plt.ylabel('log10(bs)', fontsize=5) 
    if (group == True):
        x = np.linspace(min(np.log10(bf)), max(np.log10(bf)), 100) 
        y = 0.3449 - x  
        ax.plot(x, y, color='red', linestyle='--', linewidth=0.5)
    ax.tick_params(axis='both', which='major', labelsize=4, width=0.35) 
    ax.tick_params(axis='both', which='minor', labelsize=4, width=0.35) 
    plt.xlim(xlim[0], xlim[1]) 
    plt.ylim(ylim[0], ylim[1])   
    fig.set_size_inches(2, 1.5)
    for spine in ax.spines.values(): 
        spine.set_linewidth(0.5)
    plt.show()
    return

def filter_inference_results(readfile_name, thresholds):
    inference_result_ = pd.read_csv(readfile_name)
    inference_result = np.matrix(inference_result_)[:, 1::]
    indices_infer = np.where(inference_result[:, 17].astype(float) == 0)[0]
    for index in np.arange(4):
        indices_ = np.where(inference_result[:, index+4].astype(float) == 1e-4)[0]
        indices__ = np.where(inference_result[:, index+4].astype(float) == 1e4)[0]
        indices___ = list(set(indices_) | set(indices__))
        indices_infer = list(set(indices_infer) | set(indices___))
    for index in np.arange(2):
        indices_ = np.where(np.log10(inference_result[:, index+11].astype(float)) < thresholds[0])[0]
        indices__ = np.where(np.log10(inference_result[:, index+11].astype(float)) > thresholds[1])[0]
        indices___ = list(set(indices_) | set(indices__))
        indices_infer = list(set(indices_infer) | set(indices___))
    for index in np.arange(2):
        indices_ = np.where(np.log10(inference_result[:, index+13].astype(float)) < thresholds[2])[0]
        indices__ = np.where(np.log10(inference_result[:, index+13].astype(float)) > thresholds[3])[0]
        indices___ = list(set(indices_) | set(indices__))
        indices_infer = list(set(indices_infer) | set(indices___))
    filtered_inference_result = np.delete(inference_result, indices_infer, axis = 0)
    return(filtered_inference_result)


# -*- coding: utf-8 -*-
import random, copy, math
import numpy as np
from types import SimpleNamespace
from scipy.interpolate import interp1d
from scipy.stats import poisson,norm,uniform
from scipy.special import j_roots
from scipy.special import beta as beta_fun
import seaborn as sns
import matplotlib.pyplot as plt


def SSA_coexpression(params, verbose, fig):
    """
    Generate sample data based on gene-gene interactions with bursting mechanistic dynamic model.

    Parameters
    ----------
    params
        The parameters of mechanistic model.
        params = [kon1, kon2, koff1, koff2, ksyn1, ksyn2, kdeg, rK, h1, h2, eps_, T].
    verbose
        Whether to output busrst kinetics.
        If 'burst', return with busrst kinetics.
        The priori information from the base-GRN built from scATAC-seq data, containing names and regulatory types of two genes.
    fig
        Whether to exhibit distribution images.
        If False, no distribution images are shown. If True, the images of joint distribution and marginal distributions are shown.
    
    Returns
    -------
    Sample data based on gene-gene interactions with bursting mechanistic model and its corresponding parameters.

   """
    
    kon1, kon2, koff1, koff2, ksyn1, ksyn2, kdeg = params[0: 7]  
    r_K, h1, h2, eps_ = params[7: 11]  
    T = params[11]
                             #  OFFx  ONx  OFFy  ONy   X    Y
    reactionmatrix = np.matrix([[-1,   1,   0,    0,   0,   0],
                                [ 1,  -1,   0,    0,   0,   0],
                                [ 0,   0,  -1,    1,   0,   0],
                                [ 0,   0,   1,   -1,   0,   0],
                                [ 0,   0,   0,    0,   1,   0],
                                [ 0,   0,   0,    0,   0,   1],
                                [ 0,   0,   0,    0,  -1,   0],
                                [ 0,   0,   0,    0,   0,  -1]])
    m = np.shape(reactionmatrix)[0]
    S = np.matrix([1, 0, 1, 0, 2, 2])  # initial states
    t = [0]
    tt = 0
    interval_time = []   
    bs1 = []
    bs2 = []
    kkonx = []
    kkony = []
    while (tt < T):
        if (h1 < 0):
            if (S[-1, 5] == 0): konx = kon1 * eps_
            else: konx = kon1 * (1 / (1 + 1/((r_K * S[-1, 5]) ** (-h1))) + eps_)
        else:
            konx = (h1 == 0) * kon1 * (1/2 + eps_) + (S[-1, 5] == 0) * (kon1 * (1 + eps_)) * (h1 > 0) + (S[-1, 5] > 0) * kon1 * (1 / (1 + (r_K * S[-1, 5])**h1) + eps_) * (h1 > 0) 
        if (h2 < 0):
            if (S[-1, 4] == 0): kony = kon2 * eps_
            else:  kony = kon2 * (1 / (1 + 1/((r_K * S[-1, 4]) ** (-h2))) + eps_)
        else:
            kony = (h2 == 0) * kon2 * (1/2 + eps_) + (S[-1, 4] == 0) * (kon2 * (1 + eps_)) * (h2 > 0) + (S[-1, 4] > 0) * kon2 * (1 / (1 + (r_K * S[-1, 4])**h2) + eps_) * (h2 > 0)
        PF = np.array([konx * (S[-1, 0]==1), koff1 * (S[-1, 1] == 1), kony * (S[-1, 2] == 1), koff2 * (S[-1,3] == 1), ksyn1 * (S[-1,1] == 1), ksyn2 * (S[-1,3] == 1), kdeg * S[-1, 4], kdeg * S[-1, 5]])
        p0 = np.sum(PF)
        
        r1 = random.random()
        tau = (1/p0) * np.log(1/r1)
        r2 = random.random()
        for index in np.arange(m):
            if (np.sum(PF[0:index+1]) >= r2*p0):
                next_r = index   
                break
        tt = tt + tau
        t.append(tt)
        gain = reactionmatrix[next_r, :]
        last = S[-1, :] + gain
        S = np.concatenate((S, last), axis=0)
        interval_time.append(tau)
        bs2.append(next_r == 5)
        bs1.append(next_r == 4)
        bs2.append(next_r == 5)
        kkonx.append(konx)
        kkony.append(kony)
    t = np.asarray(t)
    tq = np.arange(T-1000, T, 0.1)
    xx = np.array(S[:, 4]).flatten()
    x_interpfunc = interp1d(t, xx, 'previous')
    xx_ = x_interpfunc(tq)  
    yy = np.array(S[:, 5]).flatten()
    y_interpfunc = interp1d(t, yy, 'previous')
    yy_ = y_interpfunc(tq)
    S_stable = np.vstack((xx_, yy_))
    
    if (fig == True):
        fig, ax = plt.subplots(dpi=900)
        g = sns.jointplot(x=S_stable[0, :], y=S_stable[1, :], kind="kde", fill=True, marginal_kws=dict(fill=True), cut=0)
        g.ax_joint.set_xlim([0, np.max(S_stable[0, :])])
        g.ax_joint.set_ylim([0, np.max(S_stable[1, :])])
        g.ax_marg_x.set_xlim([0, np.max(S_stable[0, :])])
        g.ax_marg_y.set_ylim([0, np.max(S_stable[1, :])])
        g.ax_joint.set_xlabel('$X_1$')
        g.ax_joint.set_ylabel('$X_2$')    
        plt.show()
    if (verbose == 'burst'):
        D1 = np.diff(S[:, 1], axis = 0)
        D2 = np.diff(S[:, 3], axis = 0)
        index1 = np.where(D1 == 1)[0] + 1
        index2 = np.where(D2 == 1)[0] + 1
        intervaltime_1 = []
        intervaltime_2 = []
        bs_1 = []
        bs_2 = []
        nn = np.min(len(index1))
        nn0 = int(np.round(nn / 3))
        for idx in np.arange(nn0, nn): 
            intervaltime_1.append(np.sum(interval_time[index1[idx - 1]: index1[idx]]))
            bs_1.append(np.sum(bs1[index1[idx - 1]: index1[idx]]))
        nn = np.min(len(index2))
        nn0 = int(np.round(nn / 3))
        for idx in np.arange(nn0, nn): 
            intervaltime_2.append(np.sum(interval_time[index2[idx - 1]: index2[idx]]))
            bs_2.append(np.sum(bs2[index2[idx - 1]: index2[idx]]))
        mbf1 = 1 / np.mean(intervaltime_1)
        mbf2 = 1 / np.mean(intervaltime_2)
        mbs1 = np.mean(bs_1)
        mbs2 = np.mean(bs_2)
        mkonx = np.mean(kkonx[1000::])
        mkony = np.mean(kkony[1000::])
        mbf1_ = 1 / (1 / mkonx + 1 / koff1)
        mbf2_ = 1 / (1 / mkony + 1 / koff2)
        mbs1_ = ksyn1 / koff1
        mbs2_ = ksyn2 / koff2
        return SimpleNamespace(S_stable=S_stable, mbf1=mbf1, mbf2=mbf2, mbs1=mbs1, mbs2=mbs2)
    else: return (S_stable)
        
def gibbs_sample(params, m, n):
    """
    Generate sample data based on the bivaraite Poisson-Beta distribution via gibbs sampling.

    Parameters
    ----------
    params
        The parameters of statistical model.
    m
        The number of sample.
    n
        The number of iterations.
    
    Returns
    -------
    Sample data based on gene-gene interactions with bursting statistical model.

   """
    
    X = []
    Y = []
    y = [15]
    for i in range(n):
            x = sample_x(params, y[0], 1)
            y = sample_y(params, x[0], 1)
            if (i >= m+1):
                X.append(x)
                Y.append(y)
    X = np.array(X)
    Y = np.array(Y)
    X.shape = (1, len(X))
    Y.shape = (1, len(Y))
    result = np.vstack([X, Y])
    sample = np.array(result)
    return(sample)

def PoissonPMF(A,lam):
    if(max(lam) < 1e6):
        return(poisson.pmf(A,lam))
    else: return(norm.pdf(A,loc=lam,scale=math.sqrt(lam)))

def PoBePDF(A, alpha, beta, phi):    
    row,col = np.shape(A)
    prob = np.zeros((row,col))
    for i in range(row):
        aa = A[i] 
        aa = np.array(aa)
        aa.shape = (len(aa),1)
        x,w = j_roots(col, beta[i]-1, alpha[i]-1)
        GJInt = np.sum(w*PoissonPMF(aa, phi[i]*(1+x)/2), axis=1) 
        prob[i] = 1/beta_fun(alpha[i], beta[i])*2**(-float(alpha[i])-float(beta[i])+1)*GJInt
    return(prob)

def ConditionProb(params, vals): 
        row,col = np.shape(vals)
        a = params[0:2]
        b = params[2:4]
        phi = params[4:6]
        w = params[6]
        mu1 = a[0]/(a[0]+b[0])
        mu2 = a[1]/(a[1]+b[1])
        ac = copy.deepcopy(a)
        for i in range(len(a)):
            ac[i] = a[i] + 1
        uniprob = PoBePDF(vals, a, b, phi)
        uniprob1 = PoBePDF(vals, ac, b, phi)
        p_xgiveny = []
        p_ygivenx = []
        for i in range(col):
            p1 = uniprob[0,i] + w * mu1 * mu2 * (uniprob1[0,i] - uniprob[0,i]) * (uniprob1[1,i] / uniprob[1,i] - 1)
            p2 = uniprob[1,i] + w * mu1 * mu2 * (uniprob1[1,i] - uniprob[1,i]) * (uniprob1[0,i] / uniprob[0,i] - 1)
            p_xgiveny.append(p1)
            p_ygivenx.append(p2)
        px = np.array(p_xgiveny)
        py = np.array(p_ygivenx)
        px.shape = (1,col)
        py.shape = (1,col)
        result = np.vstack([px,py])
        return(result)

def sample_x(params, y, num):
    # Inverse Sampling
    sample = [0 for ss in range(num)]
    Y = [y for ss in range(200)]
    Y = np.array(Y)
    vals = np.vstack((np.arange(200), Y))
    p = ConditionProb(params, vals)[0,]
    for i in range(num):
        u = uniform.rvs(size=1)
        for j in range(200):
            index = np.arange(j+1)
            cdf = np.sum(p[index])
            if (u <= cdf):
                sample[i] = j
                break
    sample = np.array(sample)
    return(sample)

def sample_y(params, x, num): 
    sample = [0 for ss in range(num)]
    X = [x for ss in range(200)]
    X = np.array(X)
    vals = np.vstack([X, np.arange(200)])
    p = ConditionProb(params,vals)[1,]
    for i in range(num):
        u = uniform.rvs(size=1)
        for j in range(200):
            index = np.arange(j+1)
            cdf = np.sum(p[index])
            if (u <= cdf):
                sample[i] = j
                break    
    sample = np.array(sample)
    return(sample)




# -*- coding: utf-8 -*-
import os
import numpy as np
import pandas as pd
from scipy import stats
from scipy.stats import zscore, ttest_ind
from sklearn.linear_model import BayesianRidge
from scipy.stats import t
from sklearn.utils import resample
import gseapy as gp
import matplotlib.pyplot as plt
import seaborn as sns

from .._utils import _plotting as _plotting
from ..plotting import _plotting as _Plotting


def tf_tg_analysis(burst_info_file, degree_file):
    """
    Downstream analysis for gene roles in network.
    """
    burst_info_ = pd.read_csv(burst_info_file)
    burst_info = np.matrix(burst_info_)[:, 1::]
    degree_ = pd.read_csv(degree_file)
    degree = np.matrix(degree_)
    indices_tf = np.where(degree[:, 0] - degree[:, 1] < -3)[0]
    indices_tg = np.where(degree[:, 0] - degree[:, 1] > 3)[0]
    expressioninfo_tf = burst_info[indices_tf, :]
    expressioninfo_tg = burst_info[indices_tg, :]
    tf_tg_violin_plot(np.log10(expressioninfo_tf[:, 2]), np.log10(expressioninfo_tg[:, 2]))
    tf_tg_violin_plot(np.log10(expressioninfo_tf[:, 0]), np.log10(expressioninfo_tg[:, 0]))
    tf_tg_violin_plot(np.log10(expressioninfo_tf[:, 1]), np.log10(expressioninfo_tg[:, 1]))
    return

def tf_tg_violin_plot(data1, data2):
    t_stat, p_value = stats.ttest_ind(data1, data2)
    fig, ax = plt.subplots(dpi=900)
    data = [np.array(data1).flatten(), np.array(data2).flatten()]
    ax.violinplot(data, showmedians=True)
    ax.set_xticks(np.arange(1, len(data) + 1))
    ax.set_xticklabels(['tf gene', 'tg gene'])
    ax.set_ylabel('Value')
    ax.set_title('Violin Plot')  
    plt.show()
    return

def interaction_burst_regression(infer_result_filename):
    """
    Bayesian regression for gene regulatory loops and transcrptional bursting.
    """
    filtered_inference_result, unique_genename, result_unique_ = unique_infer_info(infer_result_filename, [-3.5, 2.5, -2.0, 3.0])
    # result_unique_ = [bf, bs, cv2, expression_level, overall_interaction]
    expression_level = result_unique_[:, 3]
    indices_low = np.where(np.log10(expression_level) < np.mean(np.log10(expression_level)))[0]
    indices_high = np.where(np.log10(expression_level) > np.mean(np.log10(expression_level)))[0]
    interaction_matrix_low, cmi_low = interaction_info_matrix(filtered_inference_result, unique_genename, indices_low)
    interaction_matrix_high, cmi_high = interaction_info_matrix(filtered_inference_result, unique_genename, indices_high)

    bf_results_low_ = bayes_ridge_prediction(np.log10(result_unique_[indices_low, 0]), interaction_matrix_low, cmi_low, unique_genename)
    cv2_results_low_ = bayes_ridge_prediction(np.log10(result_unique_[indices_low, 2]), interaction_matrix_low, cmi_low, unique_genename)
    bf_results_high_ = bayes_ridge_prediction(np.log10(result_unique_[indices_high, 0]), interaction_matrix_high, cmi_high, unique_genename)
    cv2_results_high_ = bayes_ridge_prediction(np.log10(result_unique_[indices_high, 2]), interaction_matrix_high, cmi_high, unique_genename)
  
    bf_results_low = bayes_ridge_prediction(result_unique_[indices_low, 0], interaction_matrix_low, cmi_low, unique_genename)
    cv2_results_low = bayes_ridge_prediction(result_unique_[indices_low, 2], interaction_matrix_low, cmi_low, unique_genename)
    bf_results_high = bayes_ridge_prediction(result_unique_[indices_high, 0], interaction_matrix_high, cmi_high, unique_genename)
    cv2_results_high = bayes_ridge_prediction(result_unique_[indices_high, 2], interaction_matrix_high, cmi_high, unique_genename)
    
    bf = np.array([bf_results_low[0][0], bf_results_low[2][0], bf_results_low[4][0], bf_results_high[0][0], bf_results_high[2][0], bf_results_high[4][0]])
    cv2 = np.array([cv2_results_low[0][0], cv2_results_low[2][0], cv2_results_low[4][0], cv2_results_high[0][0], cv2_results_high[2][0], cv2_results_high[4][0]])
    bf_sd = np.array([bf_results_low_[1][0], bf_results_low_[3][0], bf_results_low_[5][0], bf_results_high_[1][0], bf_results_high_[3][0], bf_results_high_[5][0]])
    cv2_sd = np.array([cv2_results_low_[1][0], cv2_results_low_[3][0], cv2_results_low_[5][0], cv2_results_high_[1][0], cv2_results_high_[3][0], cv2_results_high_[5][0]])
    
    _plotting.scatter_plot(np.log10(bf), np.log10(cv2), bf_sd, cv2_sd)
    return

def interaction_info_matrix(filtered_inference_result, unique_genename, indices):
    unique_genename_ = np.asarray(unique_genename)[indices]
    interaction_info_ = np.vstack([filtered_inference_result[:, np.array([0, 1, 21, 23])], filtered_inference_result[:, np.array([1, 0, 20, 22])]])
    indices = np.where(np.isin(interaction_info_[: ,0], unique_genename_))[0]
    interaction_info_ = interaction_info_[indices, :]
    interaction_info = np.matrix(sorted(interaction_info_.tolist(), key = lambda x: (x[0], x[1])))
    interaction_matrix = np.zeros([len(unique_genename_), len(unique_genename)])
    index = 0
    for genename in unique_genename_:
        indices_ = np.where(interaction_info[:, 0] == genename)[0]
        indices__ = np.where(np.isin(unique_genename, interaction_info[indices_ , 1]))[0]
        interaction_matrix[index, indices__] = np.array(interaction_info[indices_, 2].astype(float)).flatten() *  np.array(interaction_info[indices_, 3].astype(float)).flatten()
        index = index + 1
    interaction_matrix_ = np.where(interaction_matrix == 0, np.nan, interaction_matrix)
    column_cmi = np.abs(np.nanmean(interaction_matrix_, axis=0))
    return(interaction_matrix, column_cmi)
   
def bayes_ridge_prediction(y, x, w, unique_genename):
    br = BayesianRidge()
    br.fit(x, y)
    data_none = np.zeros([1, len(unique_genename)])
    data_positive = w.reshape([1, len(unique_genename)])
    data_negative = -w.reshape([1, len(unique_genename)])
    predictions1, std_dev1 = br.predict(data_positive, return_std=True)
    predictions2, std_dev2 = br.predict(data_negative, return_std=True)
    predictions3, std_dev3 = br.predict(data_none, return_std=True)
    return(predictions1, np.sqrt(std_dev1), predictions2, np.sqrt(std_dev2), predictions3, np.sqrt(std_dev3))

def unique_infer_info(infer_result_filename, thresholds):
    filtered_inference_result = _Plotting.filter_inference_results(infer_result_filename, thresholds)
    result_ = np.vstack([filtered_inference_result[:, np.array([0, 11, 13, 15, 21, 23])], filtered_inference_result[:, np.array([1, 12, 14, 16, 20, 22])]])
    sorted_result = np.matrix(sorted(result_.tolist(), key = lambda x: (x[0])))
    unique_genename = sorted(list(set(np.array(sorted_result[:, 0]).flatten().tolist())))
    result_unique_ = np.empty([len(unique_genename), 5])
    index = 0
    for genename in unique_genename:
        indices_ = np.where(sorted_result[:, 0] == genename)[0]
        if (len(indices_) == 1):
            for n in np.arange(3):
                result_unique_[index, n] = sorted_result[indices_, n+1].astype(float)
            result_unique_[index, 3] = result_unique_[index, 0] * result_unique_[index, 1]
            result_unique_[index, 4] = sorted_result[indices_, 4].astype(float) * sorted_result[indices_, 5].astype(float)
        elif (len(indices_) > 1):
            for n in np.arange(3):
                data = np.array(sorted_result[indices_, n+1].astype(float)).flatten()
                z_scores = np.abs(stats.zscore(data))
                filtered_data = np.array(data)[z_scores < 2]
                result_unique_[index, n] = np.mean(filtered_data)
            result_unique_[index, 3] = result_unique_[index, 0] * result_unique_[index, 1]
            result_unique_[index, 4] = np.sum(np.array(sorted_result[indices_, 4].astype(float)).flatten() * np.array(sorted_result[indices_, 5].astype(float)).flatten())
        index = index + 1  
    return(filtered_inference_result, unique_genename, result_unique_)

def burst_interaction_overall(burst_info_file):
    """
    Downstream analysis for gene overall regulation and transcrptional bursting.
    """
    burst_info_ = pd.read_csv(burst_info_file)
    burst_info = np.matrix(burst_info_)[:, 1::]
    expression_level = burst_info[:, 3]
    indices_low = np.where(np.log10(expression_level) < np.mean(np.log10(expression_level)))[0]
    indices_high = np.where(np.log10(expression_level) > np.mean(np.log10(expression_level)))[0]
    burst_info_low = burst_info[indices_low, :]
    burst_info_high = burst_info[indices_high, :]
    pvalue_bf_low, pvalue_bs_low, pvalue_cv2_low, pvalue_bf_high, pvalue_bs_high, pvalue_cv2_high = burst_interaction_overall_plot(burst_info_low, burst_info_high)
    return(pvalue_bf_low, pvalue_bs_low, pvalue_cv2_low, pvalue_bf_high, pvalue_bs_high, pvalue_cv2_high)

def burst_interaction_overall_plot(burstinfo1, burstinfo2):
    interaction_level1 = (burstinfo1[:, 4] - np.mean(burstinfo1[:, 4])) / np.std(burstinfo1[:, 4])
    indices_negative1 = np.where(interaction_level1 < 0)[0]
    indices_positive1 = np.where(interaction_level1 > 0)[0]
    interaction_level2 = (burstinfo2[:, 4] - np.mean(burstinfo2[:, 4])) / np.std(burstinfo2[:, 4])
    indices_negative2 = np.where(interaction_level2 < 0)[0]
    indices_positive2 = np.where(interaction_level2 > 0)[0]
    p_value1_bf, p_value1_bs, p_value1_cv2, p_value2_bf, p_value2_bs, p_value2_cv2 = _plotting.box_plot3_2(np.log10(burstinfo1[indices_positive1, 0]), np.log10(burstinfo1[indices_negative1, 0]), np.log10(burstinfo1[indices_positive1, 1]), np.log10(burstinfo1[indices_negative1, 1]), np.log10(burstinfo1[indices_positive1, 2]), np.log10(burstinfo1[indices_negative1, 2]), np.log10(burstinfo2[indices_positive2, 0]), np.log10(burstinfo2[indices_negative2, 0]), np.log10(burstinfo2[indices_positive2, 1]), np.log10(burstinfo2[indices_negative2, 1]), np.log10(burstinfo2[indices_positive2, 2]), np.log10(burstinfo2[indices_negative2, 2]))
    return(p_value1_bf, p_value1_bs, p_value1_cv2, p_value2_bf, p_value2_bs, p_value2_cv2)

def burst_info_summarize(infer_result_filename):
    inference_result__ = os.path.abspath(infer_result_filename)
    inference_result_ = pd.read_csv(inference_result__)
    inference_result = np.matrix(inference_result_)[:, 1::]
    # filter out
    indices_infer = np.where(inference_result[:, 17].astype(float) == 0)[0]
    for index in np.arange(4):
        indices_ = np.where(inference_result[:, index+4].astype(float) == 1e-4)[0]
        indices__ = np.where(inference_result[:, index+4].astype(float) == 1e4)[0]
        indices___ = list(set(indices_) | set(indices__))
        indices_infer = list(set(indices_infer) | set(indices___))
    for index in np.arange(2):
        indices_ = np.where(np.log10(inference_result[:, index+11].astype(float)) < -3.5)[0]
        indices__ = np.where(np.log10(inference_result[:, index+11].astype(float)) > 2.5)[0]
        indices___ = list(set(indices_) | set(indices__))
        indices_infer = list(set(indices_infer) | set(indices___))
    for index in np.arange(2):
        indices_ = np.where(np.log10(inference_result[:, index+13].astype(float)) < -2.0)[0]
        indices__ = np.where(np.log10(inference_result[:, index+13].astype(float)) > 3.0)[0]
        indices___ = list(set(indices_) | set(indices__))
        indices_infer = list(set(indices_infer) | set(indices___))
    filtered_inference_result = np.delete(inference_result, indices_infer, axis = 0)
    result_ = np.vstack([filtered_inference_result[:, np.array([0, 4, 6, 11, 13, 15, 21, 23])], filtered_inference_result[:, np.array([1, 5, 7, 12, 14, 16, 20, 22])]])
    sorted_result = np.matrix(sorted(result_.tolist(), key = lambda x: (x[0])))
    unique_genename = sorted(list(set(np.array(sorted_result[:, 0]).flatten().tolist())))
    result_unique_ = np.empty([len(unique_genename), 7])
    index = 0
    for genename in unique_genename:
        indices_ = np.where(sorted_result[:, 0] == genename)[0]
        if (len(indices_) == 1):
            for n in np.arange(5):
                result_unique_[index, n] = sorted_result[indices_, n+1].astype(float)
            result_unique_[index, 5] = result_unique_[index, 2] * result_unique_[index, 3]
            result_unique_[index, 6] = sorted_result[indices_, 6].astype(float) * sorted_result[indices_, 7].astype(float)
        elif (len(indices_) > 1):
            for n in np.arange(5):
                data = np.array(sorted_result[indices_, n+1].astype(float)).flatten()
                z_scores = np.abs(stats.zscore(data))
                filtered_data = np.array(data)[z_scores < 2]
                result_unique_[index, n] = np.mean(filtered_data)
            result_unique_[index, 5] = result_unique_[index, 2] * result_unique_[index, 3]
            result_unique_[index, 6] = np.sum(np.array(sorted_result[indices_, 6].astype(float)).flatten() * np.array(sorted_result[indices_, 7].astype(float)).flatten())
        index = index + 1 
    return(result_unique_)  

def affinity_burst(burst_info):
    """
    Downstream analysis for TF binding affinity and transcrptional bursting.
    """
    r = burst_info[:, 0] / burst_info[:, 1]
    bf = burst_info[:, 2]
    bs = burst_info[:, 3]
    expression_level = burst_info[:, 5]
    e, a, b, p_values_e, p_values_a, p_values_b = _plotting.fit_3d(np.log10(bf), np.log10(bs), np.log10(r))
    fig = plt.figure(figsize=(2, 1.5), dpi=300)
    ax = fig.add_subplot(111, projection='3d')
    scatter = ax.scatter(np.log10(bf), np.log10(bs), np.log10(r), c=np.log10(expression_level), s=3.5, cmap='viridis', marker='o', alpha=1)
    ax.set_xlabel('log10(bf)', fontsize=4, labelpad=-10)
    ax.set_ylabel('log10(bs)', fontsize=4, labelpad=-10)
    ax.set_zlabel('log10(r)', fontsize=4, labelpad=-10)
    ax.xaxis.label.set_position((0.5, -0.15))  
    ax.yaxis.label.set_position((-0.2, 0.5)) 
    ax.zaxis.label.set_position((-0.15, 0.5))  
    ax.tick_params(axis='x', which='major', labelsize=3, width=0.15, pad=-2) 
    ax.tick_params(axis='y', which='major', labelsize=3, width=0.15, pad=-2)  
    ax.tick_params(axis='z', which='major', labelsize=3, width=0.15, pad=-2)
    ax.w_xaxis.line.set_linewidth(0.2)
    ax.w_yaxis.line.set_linewidth(0.2)
    ax.w_zaxis.line.set_linewidth(0.2)
    plt.show()
    p_value_bf = _plotting.scatter_2d(np.log10(bf), np.log10(r), np.log10(expression_level), 'log10(bf)', 'log10(r)')
    p_value_bs = _plotting.scatter_2d(np.log10(bs), np.log10(r), np.log10(expression_level), 'log10(bs)', 'log10(r)')
    return(a, b, p_value_bf, p_value_bs)

def burst_interactionlevel_positive(readfile_name, burst_info, cv2_info):
    """
    Downstream analysis for TF binding numbers and transcrptional bursting.
    """
    degree_result_ = pd.read_csv(readfile_name)
    degree_result = np.matrix(degree_result_)
    interaction_level = zscore(burst_info[:, 6])
    indices_positive = np.where(interaction_level > 0)[0]
    indegree = np.asarray(degree_result[indices_positive, 0])
    bf = burst_info[indices_positive, 2]
    bs = burst_info[indices_positive, 3]
    expression_level = burst_info[indices_positive, 5]
    inter, a, b, p_values_inter, p_values_a, p_values_b = _plotting.fit_3d(np.log10(bf), np.log10(bs), np.log10(indegree))
    p_value_bf = _plotting.scatter_2d(np.log10(bf), np.log10(indegree), np.log10(expression_level), 'log10(bf)', 'log10(indegree)')
    p_value_bs = _plotting.scatter_2d(np.log10(bs), np.log10(indegree), np.log10(expression_level), 'log10(bs)', 'log10(indegree)')

    sorted_indegree_cv2_ = pd.read_csv(cv2_info)
    sorted_indegree_cv2 = np.array(sorted_indegree_cv2_)[:, 1::]
    x = sorted_indegree_cv2[:, 0]
    y = sorted_indegree_cv2[:, 1]
    degree = 4
    coefficients = np.polyfit(x, y, degree) 
    polynomial = np.poly1d(coefficients) 
    x_fit = np.linspace(min(x), max(x), 100)
    y_fit = polynomial(x_fit)
    n = len(y) 
    y_hat = polynomial(x) 
    residuals = y - y_hat
    mean_x = np.mean(x)
    df = n - degree - 1  
    residual_std_error = np.sqrt(np.sum(residuals**2) / df)
    t_value = t.ppf(0.95, df)
    ci = t_value * residual_std_error * np.sqrt(1/n + (x_fit - mean_x)**2 / np.sum((x - mean_x)**2))

    fig = plt.figure(figsize=(3, 1.2), dpi=300)
    ax1 = fig.add_subplot(121, projection='3d')
    scatter = ax1.scatter(np.log10(bf), np.log10(bs), np.log10(indegree), c=np.log10(expression_level), s=3, cmap='viridis', marker='o', alpha=1)
    ax1.set_xlabel('log10(bf)', fontsize=4, labelpad=-10)
    ax1.set_ylabel('log10(bs)', fontsize=4, labelpad=-10)
    ax1.set_zlabel('log10(indegree)', fontsize=4, labelpad=-10)
    ax1.xaxis.label.set_position((0.5, -0.15))  
    ax1.yaxis.label.set_position((-0.2, 0.5)) 
    ax1.zaxis.label.set_position((-0.15, 0.5))  
    ax1.tick_params(axis='x', which='major', labelsize=3, width=0.15, pad=-2) 
    ax1.tick_params(axis='y', which='major', labelsize=3, width=0.15, pad=-2)  
    ax1.tick_params(axis='z', which='major', labelsize=3, width=0.15, pad=-2)
    ax1.w_xaxis.line.set_linewidth(0.2)
    ax1.w_yaxis.line.set_linewidth(0.2)
    ax1.w_zaxis.line.set_linewidth(0.2)
    ax2 = fig.add_subplot(122) 
    ax2.scatter(x, y, s = 4)
    ax2.plot(x_fit, y_fit, linewidth=0.7)
    ax2.fill_between(x_fit, y_fit - ci, y_fit + ci, alpha=0.2)
    ax2.set_xlabel('log10(indegree)', fontsize=4)
    ax2.set_ylabel('log10(cv2)', fontsize=4)
    ax2.tick_params(axis='x', which='major', labelsize=3, width=0.08)  
    ax2.tick_params(axis='y', which='major', labelsize=3, width=0.08) 
    for spine in ax2.spines.values():
        spine.set_linewidth(0.3)
    plt.tight_layout()
    plt.show()
    t_stat_cv1, p_value_cv1 = stats.ttest_ind(sorted_indegree_cv2[0:15, 0], sorted_indegree_cv2[0:15, 1])
    t_stat_cv2, p_value_cv2 = stats.ttest_ind(sorted_indegree_cv2[14::, 0], sorted_indegree_cv2[14::, 1])
    return (a, b, p_value_bf, p_value_bs, p_value_cv1, p_value_cv2)

def correlation_burst(inference_result_Fibr_, inference_result_ESC_):
    filtered_inference_result_Fibr, unique_genename_Fibr, result_unique_Fibr = unique_infer_info(inference_result_Fibr_, [-3.5, 2.5, -2.0, 3.0])
    filtered_inference_result_ESC, unique_genename_ESC, result_unique_ESC = unique_infer_info(inference_result_ESC_, [-3.5, 2.8, -2.6, 3.2])
    unique_results = (np.array(['bf_Fiber', 'bs_Fibr', 'bf_ES', 'bs_ES'])).reshape([1, 4])
    for n in np.arange(134):
        genename = unique_genename_Fibr[n]
        indice = np.where(np.array(unique_genename_ESC) == genename)[0]
        if (len(indice) == 1):
            result = (np.array([result_unique_Fibr[n, 0], result_unique_Fibr[n, 1], result_unique_ESC[indice, 0][0], result_unique_ESC[indice, 1][0]])).reshape([1, 4])
            unique_results = np.vstack([unique_results, result])
    unique_results = np.delete(unique_results, 62, axis=0)
    p_bf, p_bs = _plotting.scatter_interval(unique_results)
    return(p_bf, p_bs)

def comparison_burst_analysis(data): 
    bf = [np.log10(data[:, 5].astype(float)), np.log10(data[:, 1].astype(float))]
    bs = [np.log10(data[:, 6].astype(float)), np.log10(data[:, 2].astype(float))]
    cv2 = [np.log10(data[:, 7].astype(float)), np.log10(data[:, 3].astype(float))]
    mean = [np.log10(data[:, 5].astype(float))+np.log10(data[:, 6].astype(float)), np.log10(data[:, 1].astype(float))+np.log10(data[:, 2].astype(float))]
    t_stat_bf, p_value_bf = stats.ttest_ind(np.log10(data[:, 1].astype(float)), np.log10(data[:, 5].astype(float)))
    t_stat_bs, p_value_bs = stats.ttest_ind(np.log10(data[:, 2].astype(float)), np.log10(data[:, 6].astype(float)))
    t_stat_cv2, p_value_cv2 = stats.ttest_ind(np.log10(data[:, 3].astype(float)), np.log10(data[:, 7].astype(float)))
    t_stat_mean, p_value_mean = stats.ttest_ind(np.log10(data[:, 1].astype(float))+np.log10(data[:, 2].astype(float)), np.log10(data[:, 5].astype(float))+np.log10(data[:, 6].astype(float)))
    _plotting.box_plot4_2(bf, bs, cv2, mean)
    return(p_value_bf, p_value_bs, p_value_cv2, p_value_mean)

def comparison_regulation_analysis(data):
    genename = data[:, 0]
    regulation = np.array(np.hstack([data[:, 8].astype(float), data[:, 4].astype(float)]))  
    genename = np.array(genename)[~np.any(regulation > 1800, axis=1)] 
    regulation = np.array(regulation)[~np.any(regulation > 1800, axis=1)] 
    genename = np.array(genename)[~np.any(regulation < 100, axis=1)] 
    regulation = np.array(regulation)[~np.any(regulation < 100, axis=1)] 
    interaction_level_DMSO = (regulation[:, 0] - np.mean(regulation[:, 0])) / np.std(regulation[:, 0])
    interaction_level_idu = (regulation[:, 1] - np.mean(regulation[:, 1])) / np.std(regulation[:, 1])
    interaction_level = np.vstack([interaction_level_idu, interaction_level_DMSO])
    dd = interaction_level[1, :] - interaction_level[0, :]
    interaction_level_ = np.vstack([interaction_level, dd])
    sorted_indices = np.argsort(interaction_level_[2])[::-1] 
    sorted_interaction_level = interaction_level_[:, sorted_indices] 
    _plotting.heatmap_regulation(sorted_interaction_level[0:2, :])
    return(genename, sorted_interaction_level, interaction_level)

def comparison_regulation_difference_analysis(sorted_interaction_level):
    _plotting.barplot_regulation(sorted_interaction_level[2, :])
    return

def go_enrichment_analysis(gene_list, gene_sets="GO_Biological_Process_2021", organism="Mouse"):
    """
    GO enrichment analysis.
    """
    results = gp.enrichr(gene_list=gene_list, gene_sets=gene_sets, organism=organism, cutoff=0.05)
    return results.res2d


def differential_tg_GO(genename, interaction_level):
    dd = np.abs(interaction_level[0, :] - interaction_level[1, :])
    columns_up = np.where((interaction_level[0, :] <= interaction_level[1, :]) & (interaction_level[1, :] - interaction_level[0, :] > np.median(dd)))[0]                 
    columns_down = np.where((interaction_level[0, :] > interaction_level[1, :]) & (interaction_level[0, :] - interaction_level[1, :] > np.median(dd)))[0] 
    go_enrichment_results_up = go_enrichment_analysis(genename[columns_up, 0].tolist())
    _plotting.visualize_go_bar(go_enrichment_results_up, "Enrichment analysis of differential up-regulated genes", "Reds_r")
    go_enrichment_results_down = go_enrichment_analysis(genename[columns_down, 0].tolist())
    _plotting.visualize_go_bar(go_enrichment_results_down, "Enrichment analysis of differential down-regulated genes", "Blues_r")
    return